from relevanceai.base import _Base
from relevanceai.vector_tools.cluster import Cluster
from relevanceai.vector_tools.dim_reduction import DimReduction
from relevanceai.vector_tools.nearest_neighbours import NearestNeighbours


class VectorTools(_Base):
    """Vector Tools Client"""

    def __init__(self, project: str, api_key: str):
        self.cluster = Cluster(project=project, api_key=api_key)
        self.dim_reduction = DimReduction(project=project, api_key=api_key)
        self.nearest_neighbours = NearestNeighbours(project=project, api_key=api_key)
        super().__init__(project, api_key)



import dash
from dash import dcc
from dash import html
from relevanceai.visualise.dash_components.components.sections import Card
from relevanceai.visualise.dash_components.components.buttons import NamedSlider


def build_control_panel(app: dash.Dash) -> html.Div:
    """
    Builds the control panel for the embedding explorer.
    """
    return html.Div(
        className="three columns",
        children=[
            Card(
                [
                    NamedSlider(
                        name="Initial PCA Dimensions",
                        short="slider-pca-dimension",
                        min=2,
                        max=3,
                        step=None,
                        val=2,
                        marks={i: str(i) for i in [2, 3]},
                    ),
                ],
                className="card-style",
            ),
        ],
    )



from dash import dcc
from dash import html


def NamedSlider(name, short, min, max, step, val, marks=None):
    if marks:
        step = None
    else:
        marks = {i: i for i in range(min, max + 1, step)}

    return html.Div(
        style={"margin": "25px 5px 30px 0px"},
        children=[
            f"{name}:",
            html.Div(
                style={"margin-left": "5px"},
                children=[
                    dcc.Slider(
                        id=short,
                        min=min,
                        max=max,
                        marks=marks,
                        step=step,
                        value=val,
                    )
                ],
            ),
        ],
    )



# -*- coding: utf-8 -*-

from doc_utils import DocUtils
from typing_extensions import Literal
from typing import List, Union, Dict, Any, Tuple, Optional
from relevanceai.vector_tools.dim_reduction import DimReduction, DimReductionBase
from relevanceai.vector_tools.cluster import Cluster, ClusterBase
from relevanceai.visualise.dash_components.app import create_dash_graph
from relevanceai.vector_tools.constants import *
from relevanceai.base import _Base
from relevanceai.api.client import BatchAPIClient
from typeguard import typechecked
from dataclasses import dataclass
import plotly.graph_objs as go
import numpy as np
import pandas as pd

pd.options.mode.chained_assignment = None


RELEVANCEAI_BLUE = "#1854FF"


@dataclass
class Projector(BatchAPIClient, _Base, DocUtils):
    """
    Projector class.

    Example:
        >>> from relevanceai import Client
        >>> project = input()
        >>> api_key = input()
        >>> client = Client(project, api_key)
        >>> client.projector.plot(
                dataset_id, vector_field, number_of_points_to_render, random_state,
                dr, dr_args, dims,
                vector_label, label_char_length,
                color_label, colour_label_char_length,
                hover_label,
                cluster, cluster_args,
                )
    """

    def __init__(self, project, api_key):
        self.project = project
        self.api_key = api_key
        super().__init__(project, api_key)

    @typechecked
    def plot(
        self,
        dataset_id: str,
        vector_field: str,
        number_of_points_to_render: int = 1000,
        # Plot rendering args
        vector_label: Union[None, str] = None,
        # Dimensionality reduction args
        dr: Union[DIM_REDUCTION, DimReductionBase] = "pca",
        dims: Literal[2, 3] = 3,
        dr_args: Union[None, Dict] = None,
        # Cluster args
        cluster: Union[CLUSTER, ClusterBase] = None,
        num_clusters: Union[None, int] = 10,
        cluster_args: Union[None, Dict] = None,
        cluster_on_dr: bool = False,
        # Decoration args
        hover_label: list = [],
        show_image: bool = False,
        label_char_length: int = 50,
        marker_size: int = 5,
        interactive: bool = False,
    ):
        """
        Dimension reduce vectors and plot with functionality to visualise different clusters and nearest neighbours

        To write your own custom dimensionality reduction, you should inherit from DimReductionBase:
        from relevanceai.visualise.dim_reduction import DimReductionBase
        class CustomDimReduction(DimReductionBase):
            def fit_transform(self, vectors):
                return np.arange(512, 2)

        Example:
            >>> from relevanceai import Client
            >>> project = input()
            >>> api_key = input()
            >>> client = Client(project, api_key)
            >>> client.projector.plot(
                    dataset_id, vector_field, number_of_points_to_render, random_state,
                    dr, dr_args, dims,
                    vector_label, label_char_length,
                    color_label, colour_label_char_length,
                    hover_label,
                    cluster, cluster_args,
                    )

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        vector_field : list
            Vector field to plot
        number_of_points_to_render: int
            Number of vector fields to plot
        vector_label: string
            Field to use as label to describe vector on plot
        dr: string
            Method of dimension reduction for vectors
        dims: int
            Number of dimensions to reduce to
        dr_args: dict
            Additional arguments for dimension reduction
        cluster: string
            Method of clustering for vectors
        num_clusters: string
            Number of clusters to create
        cluster_args: dict
            Additional arguments for clustering
        cluster_on_dr: int
            Whether to cluster on the dimension reduced or original vectors
        hover_label: list
            Additional labels to include as plot labels
        show_image: bool
            Whether vector labels are image urls
        label_char_length: int
            Maximum length of text for each hover label
        marker_size: int
            Marker size of the plot
        interactive: bool
            Whether to include interactive features including nearest neighbours

        """

        # Check vector field
        self._is_valid_vector_name(dataset_id, vector_field)

        # Check vector label field
        if vector_label is None:
            self.logger.warning("A vector_label has not been specified.")
        else:
            self._is_valid_label_name(dataset_id, vector_label)

        # Check hover label field
        [self._is_valid_label_name(dataset_id, label) for label in hover_label]

        docs = self.get_documents(
            dataset_id,
            number_of_documents=number_of_points_to_render,
            batch_size=1000,
            select_fields=["_id", vector_field, vector_label] + hover_label,
        )
        docs = self._remove_empty_vector_fields(docs, vector_field)

        return self.plot_from_docs(
            docs,
            vector_field=vector_field,
            vector_label=vector_label,
            label_char_length=label_char_length,
            dr=dr,
            dims=dims,
            dr_args=dr_args,
            cluster=cluster,
            num_clusters=num_clusters,
            cluster_args=cluster_args,
            cluster_on_dr=cluster_on_dr,
            hover_label=hover_label,
            show_image=show_image,
            marker_size=marker_size,
            dataset_name=dataset_id,
            interactive=interactive,
        )

    def plot_from_docs(
        self,
        docs: List[Dict],
        vector_field: str,
        # Plot rendering args
        vector_label: Union[None, str] = None,
        # Dimensionality reduction args
        dr: Union[DIM_REDUCTION, DimReductionBase] = "pca",
        dims: Literal[2, 3] = 3,
        dr_args: Union[None, Dict] = None,
        # Cluster args
        cluster: Union[CLUSTER, ClusterBase] = None,
        num_clusters: Union[None, int] = 10,
        cluster_args: Union[None, Dict] = None,
        cluster_on_dr: bool = False,
        # Decoration args
        hover_label: list = [],
        show_image: bool = False,
        label_char_length: int = 50,
        marker_size: int = 5,
        dataset_name: Union[None, str] = None,
        interactive: bool = False,
    ):

        # Adjust vector label
        if show_image is False:
            self.set_field_across_documents(
                vector_label,
                [i[vector_label][:label_char_length] + "..." for i in docs],
                docs,
            )

        # Dimension reduce vectors
        vectors = np.array(self.get_field_across_documents(vector_field, docs))
        vectors_dr = DimReduction.dim_reduce(
            vectors=vectors, dr=dr, dr_args=dr_args, dims=dims
        )
        points = {"x": vectors_dr[:, 0], "y": vectors_dr[:, 1]}
        if dims == 3:
            points["z"] = vectors_dr[:, 2]

        embedding_df = pd.DataFrame(points)
        embedding_df = pd.concat([embedding_df, pd.DataFrame(docs)], axis=1)

        # Set hover labels
        hover_label = ["_id", vector_label] + hover_label

        # Cluster vectors
        if cluster:
            if cluster_on_dr:
                cluster_vec = vectors_dr
            else:
                cluster_vec = vectors

            cluster_labels = Cluster.cluster(
                vectors=cluster_vec,
                cluster=cluster,
                cluster_args=cluster_args,
                k=num_clusters,
            )
            embedding_df["cluster_labels"] = cluster_labels
            hover_label = hover_label + ["cluster_labels"]

        embedding_df.index = embedding_df["_id"]

        # Generate plot title
        plot_title = self._generate_plot_title(
            dims,
            dataset_name,
            len(embedding_df),
            vector_field,
            vector_label,
            label_char_length,
        )

        plot_data = self._generate_plot_data(
            embedding_df=embedding_df,
            hover_label=hover_label,
            dims=dims,
            marker_size=marker_size,
            cluster=cluster,
            label_char_length=label_char_length,
        )

        layout = self._generate_layout(plot_title=plot_title)

        create_dash_graph(
            plot_data=plot_data,
            layout=layout,
            show_image=show_image,
            docs=docs,
            vector_label=vector_label,
            vector_field=vector_field,
            interactive=interactive,
        )
        return

    def _generate_plot_data(
        self,
        embedding_df: pd.DataFrame,
        hover_label: List[Optional[str]],
        dims: int,
        marker_size: int,
        cluster: Union[
            Literal["kmeans"],
            Literal["kmedoids"],
            Literal["hdbscan"],
            ClusterBase,
            None,
        ],
        label_char_length: int,
    ):
        """ """

        if cluster:
            data = []
            groups = embedding_df.groupby("cluster_labels")
            for idx, val in groups:
                data.append(
                    self._generate_plot_info(
                        embedding_df=val,
                        hover_label=hover_label,
                        dims=dims,
                        marker_size=marker_size,
                        label_char_length=label_char_length,
                    )
                )

        else:
            data = []
            data.append(
                self._generate_plot_info(
                    embedding_df=embedding_df,
                    hover_label=hover_label,
                    dims=dims,
                    marker_size=marker_size,
                    label_char_length=label_char_length,
                )
            )

        return data

    def _generate_layout(self, plot_title):

        axes_3d = {
            "title": "",
            "backgroundcolor": "#ffffff",
            "showgrid": False,
            "showticklabels": False,
        }

        axes_2d = {"title": "", "visible": False, "showticklabels": False}

        layout = go.Layout(
            margin={"l": 0, "r": 0, "b": 0, "t": 0},
            scene={"xaxis": axes_3d, "yaxis": axes_3d, "zaxis": axes_3d},
            title={
                "text": plot_title,
                "y": 0.1,
                "x": 0.1,
                "xanchor": "left",
                "yanchor": "bottom",
                "font": {"size": 10},
            },
            plot_bgcolor="#FFF",
            xaxis=axes_2d,
            yaxis=axes_2d,
        )

        return layout

    def _generate_plot_info(
        self, embedding_df, hover_label, dims, marker_size, label_char_length
    ):

        custom_data, hovertemplate = self._generate_hover_template(
            df=embedding_df,
            dims=dims,
            hover_label=hover_label,
            label_char_length=label_char_length,
        )

        scatter_args = {
            "x": embedding_df["x"],
            "y": embedding_df["y"],
            "showlegend": False,
            "mode": "markers",
            "marker": {"size": marker_size, "symbol": "circle", "opacity": 0.75},
            "customdata": custom_data,
            "hovertemplate": hovertemplate,
        }

        if dims == 2:
            scatter = go.Scatter(**scatter_args)

        else:
            scatter_args["z"] = embedding_df["z"]
            scatter = go.Scatter3d(**scatter_args)

        return scatter

    def _generate_hover_template(
        self, df: pd.DataFrame, dims: int, hover_label: list, label_char_length: int
    ):
        """
        Generating hover template
        """
        custom_data = df[hover_label]
        custom_data = custom_data.loc[:, ~custom_data.columns.duplicated()]

        for label in hover_label:
            try:
                if label != "_id":
                    custom_data.loc[:, label] = [
                        i[:label_char_length] + "..." for i in custom_data.loc[:, label]
                    ]
            except:
                pass

        custom_data_hover = [
            f"{c}: %{{customdata[{i}]}}" for i, c in enumerate(hover_label)
        ]

        if dims == 2:
            coord_info = "X: %{x}   Y: %{y}"
        else:
            coord_info = "X: %{x}   Y: %{y}   Z: %{z}"

        hovertemplate = (
            "<br>".join(
                [
                    coord_info,
                ]
                + custom_data_hover
            )
            + "<extra></extra>"
        )

        return custom_data, hovertemplate

    def _generate_plot_title(
        self,
        dims,
        dataset_name,
        number_of_points,
        vector_field,
        vector_label,
        label_char_length,
    ):
        title = "</b>"
        title += f"{dims}D Embedding Projector Plot<br>"
        if dataset_name:
            title += f"Dataset Name: {dataset_name}<br>"
        title += f"Points: {number_of_points} points<br>"
        title += f"Vector Field: {vector_field}<br>"
        title += f"Vector Label: {vector_label}  Char Length: {label_char_length}<br>"
        title += "</b>"
        return title

    def _is_valid_vector_name(self, dataset_id, vector_name: str) -> bool:
        """
        Check vector field name is valid
        """
        vector_fields = self.get_vector_fields(dataset_id)
        schema = self.datasets.schema(dataset_id)
        if vector_name in schema.keys():
            if vector_name in vector_fields:
                return True
            else:
                raise ValueError(f"{vector_name} is not a valid vector name")
        else:
            raise ValueError(f"{vector_name} is not in the {dataset_id} schema")

    def _is_valid_label_name(self, dataset_id, label_name: str) -> bool:
        """
        Check vector label name is valid. Checks that it is either numeric or text
        """
        schema = self.datasets.schema(dataset_id)
        if label_name == "_id":
            return True
        if label_name in list(schema.keys()):
            if schema[label_name] in ["numeric", "text"]:
                return True
            else:
                raise ValueError(f"{label_name} is not a valid label name")
        else:
            raise ValueError(f"{label_name} is not in the {dataset_id} schema")

    def _remove_empty_vector_fields(self, docs, vector_field: str) -> List[Dict]:
        """
        Remove documents with empty vector fields
        """
        return [d for d in docs if d.get(vector_field)]



import sys
import time

from relevanceai.config import CONFIG
from relevanceai.transport import Transport
from relevanceai.logger import LoguruLogger


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


class _Base(Transport, LoguruLogger):
    """Base class for all relevanceai client utilities"""

    def __init__(self, project: str, api_key: str):
        self.project = project
        self.api_key = api_key
        self.config = CONFIG
        # Initialize logger
        super().__init__()



import dash
from dash import dcc
from dash import html


def build_header(app: dash.Dash) -> html.Div:
    """
    Builds the header of the app.
    """
    return html.Div(
        className="row header",
        id="app-header",
        style={"background-color": "#0F172A", "margin-bottom": "16px"},
        children=[
            # html.Div(
            #     [
            #         html.H3(
            #             "Vector Projector",
            #             className="header_title",
            #             id="app-title",
            #         )
            #     ],
            #     className="nine columns header_title_container",
            # ),
        ],
    )



import dash
from dash import dcc
from dash import html
from relevanceai.visualise.dash_components.components.sections import Card


def build_display_panel(app: dash.Dash, show_image: bool = True) -> html.Div:
    """
    Builds the display panel.
    """
    if show_image:
        children = Card(
            style={"width": "98vh", "textwd-align": "center"},
            children=[html.Div(id="div-plot-image-neighbours")],
        )
    else:
        children = Card(
            style={"width": "98vh", "textwd-align": "center"},
            children=[html.Div(id="div-plot-text-neighbours")],
        )
    return html.Div(
        className="three columns",
        id="display-panel",
        children=[children],
    )



"""API Client
"""
from relevanceai.base import _Base
from relevanceai.api.endpoints.admin import AdminClient
from relevanceai.api.endpoints.datasets import DatasetsClient
from relevanceai.api.endpoints.services import ServicesClient
from relevanceai.datasets import ExampleDatasets


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


class APIClient(_Base):
    """API Client"""

    def __init__(self, project: str, api_key: str):
        self.datasets = DatasetsClient(project=project, api_key=api_key)
        self.services = ServicesClient(project=project, api_key=api_key)
        self.example_datasets = ExampleDatasets()
        self.admin = AdminClient(project=project, api_key=api_key)
        super().__init__(project, api_key)



"""Tagger services
"""
from relevanceai.base import _Base


class TaggerClient(_Base):
    def __init__(self, project, api_key):
        self.project = project
        self.api_key = api_key
        super().__init__(project, api_key)

    def tag(
        self,
        data: str,
        tag_dataset_id: str,
        encoder: str,
        tag_field: str = None,
        approximation_depth: int = 0,
        sum_fields: bool = True,
        page_size: int = 20,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: list = [],
        min_score: float = 0,
        include_search_relevance: bool = False,
        search_relevance_cutoff_aggressiveness: int = 1,
        asc: bool = False,
        include_score: bool = False,
    ):

        """
        Tag documents or vectors

        Parameters
        ----------
        data : string
            Image Url or text or any data suited for the encoder
        tag_dataset_id : string
            Name of the dataset you want to tag
        encoder: string
            Which encoder to use.
        tag_field: string
            The field used to tag in a dataset. If None, automatically uses the one stated in the encoder.
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        include_search_relevance: bool
            Whether to calculate a search_relevance cutoff score to flag relevant and less relevant results
        search_relevance_cutoff_aggressiveness: int
            How aggressive the search_relevance cutoff score is (higher value the less results will be relevant)
        asc: bool
            Whether to sort results by ascending or descending order
        include_score: bool
            Whether to include score
        """

        return self.make_http_request(
            f"/services/tagger/tag",
            method="POST",
            parameters={
                "data": data,
                "tag_dataset_id": tag_dataset_id,
                "encoder": encoder,
                "tag_field": tag_field,
                "approximation_depth": approximation_depth,
                "sum_fields": sum_fields,
                "page_size": page_size,
                "page": page,
                "similarity_metric": similarity_metric,
                "filters": filters,
                "min_score": min_score,
                "include_search_relevance": include_search_relevance,
                "search_relevance_cutoff_aggressiveness": search_relevance_cutoff_aggressiveness,
                "asc": asc,
                "include_score": include_score,
            },
        )

    def diversity(
        self,
        data: str,
        tag_dataset_id: str,
        encoder: str,
        cluster_vector_field: str,
        n_clusters: int,
        tag_field: str = None,
        approximation_depth: int = 0,
        sum_fields: bool = True,
        page_size: int = 20,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: list = [],
        min_score: float = 0,
        include_search_relevance: bool = False,
        search_relevance_cutoff_aggressiveness: int = 1,
        asc: bool = False,
        include_score: bool = False,
        n_init: int = 5,
        n_iter: int = 10,
    ):

        """
        Tagging and then clustering the tags and returning one from each cluster (starting from the closest tag)

        Parameters
        ----------
        data : string
            Image Url or text or any data suited for the encoder
        tag_dataset_id : string
            Name of the dataset you want to tag
        encoder: string
            Which encoder to use.
        cluster_vector_field: str
            The field to cluster on.
        n_clusters: int
            Number of clusters to be specified.
        tag_field: string
            The field used to tag in a dataset. If None, automatically uses the one stated in the encoder.
        approximation_depth: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate.
        sum_fields : bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        min_score: int
            Minimum score for similarity metric
        include_search_relevance: bool
            Whether to calculate a search_relevance cutoff score to flag relevant and less relevant results
        search_relevance_cutoff_aggressiveness: int
            How aggressive the search_relevance cutoff score is (higher value the less results will be relevant)
        asc: bool
            Whether to sort results by ascending or descending order
        include_score: bool
            Whether to include score
        n_init: int
            Number of runs to run with different centroid seeds
        n_iter: int
            Number of iterations in each run
        """

        return self.make_http_request(
            f"/services/tagger/diversity",
            method="POST",
            parameters={
                "data": data,
                "tag_dataset_id": tag_dataset_id,
                "encoder": encoder,
                "tag_field": tag_field,
                "approximation_depth": approximation_depth,
                "sum_fields": sum_fields,
                "page_size": page_size,
                "page": page,
                "similarity_metric": similarity_metric,
                "filters": filters,
                "min_score": min_score,
                "include_search_relevance": include_search_relevance,
                "search_relevance_cutoff_aggressiveness": search_relevance_cutoff_aggressiveness,
                "asc": asc,
                "include_score": include_score,
                "cluster_vector_field": cluster_vector_field,
                "n_clusters": n_clusters,
                "n_init": n_init,
                "n_iter": n_iter,
            },
        )



# -*- coding: utf-8 -*-

from abc import abstractmethod
import numpy as np
import warnings

from typing import List, Union, Dict, Any, Optional
from doc_utils import DocUtils
from joblib.memory import Memory
from relevanceai.api.client import BatchAPIClient
from relevanceai.logger import LoguruLogger
from relevanceai.vector_tools.constants import CLUSTER, CLUSTER_DEFAULT_ARGS
from relevanceai.errors import ClusteringResultsAlreadyExistsError


class ClusterBase(LoguruLogger, DocUtils):
    def __call__(self, *args, **kwargs):
        return self.fit_transform(*args, **kwargs)

    @abstractmethod
    def fit_transform(self, vectors):
        """ """
        raise NotImplementedError

    def _concat_vectors_from_list(self, list_of_vectors: list):
        """Concatenate 2 vectors together in a pairwise fashion"""
        return [np.concatenate(x) for x in list_of_vectors]

    def fit_documents(
        self,
        vector_fields: list,
        docs: list,
        alias: str = "default",
        cluster_field: str = "_cluster_",
        return_only_clusters: bool = True,
        inplace: bool = True,
    ):
        """
        Train clustering algorithm on documents and then store the labels
        inside the documents.

        Parameters
        -----------
        vector_field: list
            The vector field of the documents
        docs: list
            List of documents to run clustering on
        alias: str
            What the clusters can be called
        cluster_field: str
            What the cluster fields should be called
        return_only_clusters: bool
            If True, return only clusters, otherwise returns the original document
        inplace: bool
            If True, the documents are edited inplace otherwise, a copy is made first
        kwargs: dict
            Any other keyword argument will go directly into the clustering algorithm

        """
        if len(vector_fields) == 1:
            # filtering out entries not containing the specified vector
            docs = list(filter(DocUtils.list_doc_fields, docs))
            vectors = self.get_field_across_documents(
                vector_fields[0], docs, missing_treatment="skip"
            )
        else:
            # In multifield clusering, we get all the vectors in each document
            # (skip if they are missing any of the vectors)
            # Then run clustering on the result
            docs = list(self.filter_docs_for_fields(vector_fields, docs))
            all_vectors = self.get_fields_across_documents(
                vector_fields, docs, missing_treatment="skip_if_any_missing"
            )
            vectors = self._concat_vectors_from_list(all_vectors)

        cluster_labels = self.fit_transform(vectors)

        # Label the clusters
        cluster_labels = self._label_clusters(cluster_labels)

        if inplace:
            self.set_field_across_documents(
                f"{cluster_field}.{'.'.join(vector_fields)}.{alias}",
                cluster_labels,
                docs,
            )
            if return_only_clusters:
                return [
                    {"_id": d.get("_id"), cluster_field: d.get(cluster_field)}
                    for d in docs
                ]
            return docs

        new_docs = docs.copy()

        self.set_field_across_documents(
            f"{cluster_field}.{vector_fields[0]}.{alias}", cluster_labels, new_docs
        )

        if return_only_clusters:
            return [
                {"_id": d.get("_id"), cluster_field: d.get(cluster_field)} for d in docs
            ]
        return docs

    def to_metadata(self):
        """You can also store the metadata of this clustering algorithm"""
        raise NotImplementedError

    @property
    def metadata(self):
        return self.to_metadata()

    def _label_cluster(self, label: Union[int, str]):
        if isinstance(label, (int, float)):
            return "Cluster-" + str(label)
        return str(label)

    def _label_clusters(self, labels):
        return [self._label_cluster(x) for x in labels]


class CentroidCluster(ClusterBase):
    def __call__(self, *args, **kwargs):
        return self.fit_transform(*args, **kwargs)

    @abstractmethod
    def fit_transform(self, vectors):
        raise NotImplementedError

    @abstractmethod
    def get_centers(self) -> Union[np.ndarray, List[list]]:
        """Get centers for the centroid-based clusters"""
        raise NotImplementedError

    def get_centroid_docs(self) -> List:
        """Get the centroid documents to store."""
        self.centers = self.get_centers()
        if isinstance(self.centers, np.ndarray):
            self.centers = self.centers.tolist()
        return [
            {"_id": f"cluster_{i}", "centroid_vector_": self.centers[i]}
            for i in range(len(self.centers))
        ]


class DensityCluster(ClusterBase):
    def __call__(self, *args, **kwargs):
        return self.fit_transform(*args, **kwargs)

    def fit_transform(self, vectors):
        raise NotImplementedError


class MiniBatchKMeans(CentroidCluster):
    def __init__(
        self,
        k: Union[None, int] = 10,
        init: str = "k-means++",
        verbose: bool = False,
        compute_labels: bool = True,
        max_no_improvement: int = 2,
    ):
        """
        Kmeans Centroid Clustering

        Parameters
        ------------
        k: int
            The number of clusters
        init: str
            The optional parameter to be clustering
        verbose: bool
            If True, will print what is happening
        compute_labels: bool
            If True, computes the labels of the cluster
        max_no_improvement: int
            The maximum number of improvemnets
        """
        self.k = k
        self.init = init
        self.verbose = verbose
        self.compute_labels = compute_labels
        self.max_no_improvement = max_no_improvement

    def _init_model(self):
        from sklearn.cluster import MiniBatchKMeans

        self.km = MiniBatchKMeans(
            n_clusters=self.k,
            init=self.init,
            verbose=self.verbose,
            compute_labels=self.compute_labels,
            max_no_improvement=self.max_no_improvement,
        )
        return

    def fit_transform(self, vectors: Union[np.ndarray, List]):
        """
        Fit and transform transform the vectors
        """
        if not hasattr(self, "km"):
            self._init_model()
        self.km.fit(vectors)
        cluster_labels = self.km.labels_.tolist()
        # cluster_centroids = km.cluster_centers_
        return cluster_labels

    def get_centers(self):
        """Returns centroids of clusters"""
        return [list(i) for i in self.km.cluster_centers_]

    def to_metadata(self):
        """Editing the metadata of the function"""
        return {
            "k": self.k,
            "init": self.init,
            "verbose": self.verbose,
            "compute_labels": self.compute_labels,
            "max_no_improvement": self.max_no_improvement,
            "number_of_clusters": self.k,
        }


# class KMedoids(CentroidCluster):
#     def fit_transform(self,
#         vectors: np.ndarray,
#         cluster_args: Optional[Dict[Any, Any]] = CLUSTER_DEFAULT_ARGS['kmedoids'],
#         k: Union[None, int] = 10,
#     ) -> np.ndarray:
#         try:
#             from sklearn_extra.cluster import KMedoids
#         except ModuleNotFoundError as e:
#             raise ModuleNotFoundError(
#                 f"{e}\nInstall umap\n \
#                 pip install -U relevanceai[kmedoids]"
#             )
#         self.logger.debug(f"{cluster_args}")
#         km = KMedoids(n_clusters=k, **cluster_args).fit(vectors)
#         cluster_labels = km.labels_
#         # cluster_centroids = km.cluster_centers_
#         return cluster_labels


class KMeans(MiniBatchKMeans):
    def __init__(
        self,
        k=10,
        init="k-means++",
        n_init=10,
        max_iter=300,
        tol=1e-4,
        verbose=0,
        random_state=None,
        copy_x=True,
        algorithm="auto",
    ):
        self.init = init
        self.n_init = n_init
        self.max_iter = max_iter
        self.tol = tol
        self.verbose = verbose
        self.random_state = random_state
        self.copy_x = copy_x
        self.algorithm = algorithm
        self.n_clusters = k

    def _init_model(self):
        from sklearn.cluster import KMeans

        self.km = KMeans(
            n_clusters=self.n_clusters,
            init=self.init,
            verbose=self.verbose,
            max_iter=self.max_iter,
            tol=self.tol,
            random_state=self.random_state,
            copy_x=self.copy_x,
            algorithm=self.algorithm,
        )
        return

    def to_metadata(self):
        """Editing the metadata of the function"""
        return {
            "n_clusters": self.n_clusters,
            "init": self.init,
            "max_iter": self.max_iter,
            "tol": self.tol,
            "random_state": self.random_state,
            "copy_x": self.copy_x,
            "algorithm": self.algorithm,
        }


class HDBSCANClusterer(DensityCluster):
    def __init__(
        self,
        algorithm: str = "best",
        alpha: float = 1.0,
        approx_min_span_tree: bool = True,
        gen_min_span_tree: bool = False,
        leaf_size: int = 40,
        memory=Memory(cachedir=None),
        metric: str = "euclidean",
        min_samples=None,
        p=None,
        min_cluster_size: Union[None, int] = 10,
    ):
        self.algorithm = algorithm
        self.alpha = alpha
        self.approx_min_span_tree = approx_min_span_tree
        self.gen_min_span_tree = gen_min_span_tree
        self.leaf_size = leaf_size
        self.memory = memory
        self.metric = metric
        self.min_samples = min_samples
        self.p = p
        self.min_cluster_size = min_cluster_size

    def fit_transform(
        self,
        vectors: np.ndarray,
    ) -> np.ndarray:
        try:
            from hdbscan import HDBSCAN
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                f"{e}\nInstall hdbscan\n \
                pip install -U relevanceai[hdbscan]"
            )
        hdbscan = HDBSCAN(
            min_cluster_size=self.min_cluster_size,
            algorithm=self.algorithm,
            approx_min_span_tree=self.approx_min_span_tree,
            gen_min_span_tree=self.gen_min_span_tree,
            leaf_size=self.leaf_size,
            memory=self.memory,
            metric=self.metric,
            min_samples=self.min_samples,
            p=self.p,
        ).fit(vectors)
        cluster_labels = hdbscan.labels_
        return cluster_labels


class Cluster(BatchAPIClient, ClusterBase):
    def __init__(self, project, api_key):
        self.project = project
        self.api_key = api_key
        super().__init__(project, api_key)

    @staticmethod
    def _choose_k(vectors: np.ndarray):
        """ "
        Choose k clusters
        """
        # Partitioning methods
        # if check_type(self.cluster, CLUSTER_NUMERIC):
        """
        Scaled_inertia = inertia(k)/inertia(k=1) + (a * K)
        where a is penalty factor of num_clusters
        """
        return 10

    @staticmethod
    def cluster(
        vectors: np.ndarray,
        cluster: Union[CLUSTER, ClusterBase],
        cluster_args: Union[None, dict],
        k: Union[None, int] = None,
    ) -> np.ndarray:
        """
        Cluster vectors
        """
        if isinstance(cluster, str):
            if cluster_args is None:
                cluster_args = CLUSTER_DEFAULT_ARGS[cluster]
            if cluster in ["kmeans", "kmedoids"]:
                if (k is None and cluster_args is None) or (
                    "n_clusters" not in cluster_args.keys()
                ):
                    k = Cluster._choose_k(vectors)
                if cluster == "kmeans":
                    return KMeans(k=k, **cluster_args).fit_transform(vectors=vectors)
                elif cluster == "kmedoids":
                    raise NotImplementedError
                    # return KMedioids().fit_transform(vectors=vectors, cluster_args=cluster_args)
            elif cluster == "hdbscan":
                return HDBSCANClusterer(**cluster_args).fit_transform(vectors=vectors)

        elif isinstance(cluster, ClusterBase):
            return cluster().fit_transform(vectors=vectors, cluster_args=cluster_args)
        raise ValueError("Not valid cluster input.")

    def kmeans_cluster(
        self,
        dataset_id: str,
        vector_fields: list,
        filters: List = [],
        k: Union[None, int] = 10,
        init: str = "k-means++",
        n_init: int = 10,
        max_iter: int = 300,
        tol: float = 1e-4,
        verbose: bool = True,
        random_state: Optional[int] = None,
        copy_x: bool = True,
        algorithm: str = "auto",
        alias: str = None,
        cluster_field: str = "_cluster_",
        update_documents_chunksize: int = 50,
        overwrite: bool = False,
    ):
        """
        This function performs all the steps required for Kmeans clustering:
        1- Loads the data
        2- Clusters the data
        3- Updates the data with clustering info
        4- Adds the centroid to the hidden centroid collection

        Parameters
        ----------
        dataset_id : string
            name of the dataser
        vector_fields : list
            a list containing the vector field to be used for clustering
        filters : list
            a list to filter documents of the dataset,
        k : int
            K in Kmeans
        init : string
            "k-means++" -> Kmeans algorithm parameter
        n_init : int
            number of reinitialization for the kmeans algorithm
        max_iter : int
            max iteration in the kmeans algorithm
        tol : int
            tol in the kmeans algorithm
        verbose : bool
            True by default
        random_state = None
            None by default -> Kmeans algorithm parameter
        copy_x : bool
            True bydefault
        algorithm : string
            "auto" by default
        alias : string
            "kmeans", string to be used in naming of the field showing the clustering results
        cluster_field: string
            "_cluster_", string to name the main cluster field
        overwrite : bool
            False by default, To overwite an existing clusering result

        Example
        -------------

        >>> client.vector_tools.cluster.kmeans_cluster(
            dataset_id="sample_dataset",
            vector_fields=["sample_1_vector_"] # Only 1 vector field is supported for now
        )
        """
        if alias is None:
            alias = "kmeans_" + str(k)

        EXPECTED_CLUSTER_OUTFIELD = ".".join([cluster_field, vector_fields[0], alias])
        if (
            EXPECTED_CLUSTER_OUTFIELD in self.datasets.schema(dataset_id)
            and not overwrite
        ):
            raise ClusteringResultsAlreadyExistsError(EXPECTED_CLUSTER_OUTFIELD)

        filters = filters + [
            {
                "field": vector_fields[0],
                "filter_type": "exists",
                "condition": ">=",
                "condition_value": " ",
            }
        ]
        # load the documents
        warnings.warn(
            "Retrieving documents... This can take a while if the dataset is large."
        )
        docs = self.get_all_documents(
            dataset_id=dataset_id, filters=filters, select_fields=vector_fields
        )

        # Cluster
        clusterer = KMeans(
            k=k,
            init=init,
            n_init=n_init,
            max_iter=max_iter,
            tol=tol,
            verbose=verbose,
            random_state=random_state,
            copy_x=copy_x,
            algorithm=algorithm,
        )
        clustered_docs = clusterer.fit_documents(
            vector_fields,
            docs,
            alias=alias,
            cluster_field=cluster_field,
            return_only_clusters=True,
            inplace=False,
        )

        # Updating the db
        results = self.update_documents(
            dataset_id, clustered_docs, chunksize=update_documents_chunksize
        )
        self.logger.info(results)

        # Update the centroid collection
        centers = clusterer.get_centroid_docs()
        results = self.services.cluster.centroids.insert(
            dataset_id=dataset_id,
            cluster_centers=centers,
            vector_field=vector_fields[0],
            alias=alias,
        )
        self.logger.info(results)

        return centers

    def hdbscan_cluster(
        self,
        dataset_id: str,
        vector_fields: list,
        filters: List = [],
        algorithm: str = "best",
        alpha: float = 1.0,
        approx_min_span_tree: bool = True,
        gen_min_span_tree: bool = False,
        leaf_size: int = 40,
        memory=Memory(cachedir=None),
        metric: str = "euclidean",
        min_samples=None,
        p=None,
        min_cluster_size: Union[None, int] = 10,
        alias: str = "hdbscan",
        cluster_field: str = "_cluster_",
        update_documents_chunksize: int = 50,
        overwrite: bool = False,
    ):
        """
        This function performs all the steps required for hdbscan clustering:
        1- Loads the data
        2- Clusters the data
        3- Updates the data with clustering info
        4- Adds the centroid to the hidden centroid collection

        Parameters
        ----------
        dataset_id : string
            name of the dataser
        vector_fields : list
            a list containing the vector field to be used for clustering
        filters : list
            a list to filter documents of the dataset
        algorithm : str
            hdbscan configuration parameter default to "best"
        alpha: float
            hdbscan configuration parameter default to 1.0
        approx_min_span_tree: bool
            hdbscan configuration parameter default to True
        gen_min_span_tree: bool
            hdbscan configuration parameter default to False
        leaf_size: int
            hdbscan configuration parameter default to 40
        memory = Memory(cachedir=None)
            hdbscan configuration parameter on memory management
        metric: str = "euclidean"
            hdbscan configuration parameter default to "euclidean"
        min_samples = None
            hdbscan configuration parameter default to None
        p = None
            hdbscan configuration parameter default to None
        min_cluster_size:
            minimum cluster size, 10 by default
        alias : string
            "hdbscan", string to be used in naming of the field showing the clustering results
        cluster_field: string
            "_cluster_", string to name the main cluster field
        overwrite : bool
            False by default, To overwite an existing clusering result

        Example
        -------------

        >>> client.vector_tools.cluster.hdbscan_cluster(
            dataset_id="sample_dataset",
            vector_fields=["sample_1_vector_"] # Only 1 vector field is supported for now
        )
        """

        if (
            ".".join([cluster_field, vector_fields[0], alias])
            in self.datasets.schema(dataset_id)
            and not overwrite
        ):
            raise ClusteringResultsAlreadyExistsError(
                ".".join([cluster_field, vector_fields[0], alias])
            )
        # load the documents
        docs = self.get_all_documents(
            dataset_id=dataset_id, filters=filters, select_fields=vector_fields
        )

        # get vectors
        if len(vector_fields) > 1:
            raise ValueError(
                "We currently do not support more than 1 vector field yet. This will be supported in the future."
            )

        # Cluster
        clusterer = HDBSCANClusterer(
            algorithm=algorithm,
            alpha=alpha,
            approx_min_span_tree=approx_min_span_tree,
            gen_min_span_tree=gen_min_span_tree,
            leaf_size=leaf_size,
            memory=memory,
            metric=metric,
            min_samples=min_samples,
            p=p,
            min_cluster_size=min_cluster_size,
        )
        clustered_docs = clusterer.fit_documents(
            vector_fields, docs, alias=alias, return_only_clusters=True
        )

        # Updating the db
        # formatted_clustered_docs = [
        #     {cluster_field:{vector_fields[0]:{alias:res}},
        #     '_id':docs[i]['_id']}
        #     for i,res in enumerate(clustered_docs)]
        results = self.update_documents(
            dataset_id, clustered_docs, chunksize=update_documents_chunksize
        )
        self.logger.info(results)
        return clustered_docs



from os import error
import sys
from typing import Callable
from loguru import logger as loguru_logger
from abc import abstractmethod
from relevanceai.config import CONFIG


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


class AbstractLogger:
    """Base Logging Instance"""

    info: Callable
    error: Callable
    success: Callable
    debug: Callable
    warning: Callable
    critical: Callable
    warn: Callable
    # @property
    # @abstractmethod
    # def logger(self):
    #     raise NotImplementedError


class LoguruLogger(AbstractLogger):
    """Using verbose loguru as base logger for now"""

    # Add Logging

    def __init__(self):
        self._init_logger()

    @property
    def logger(self):
        self._init_logger()
        return self._logger

    def _init_logger(self):
        self.config = CONFIG
        logging_level = self.config.get_option("logging.logging_level")
        log_to_file = str2bool(self.config.get_option("logging.log_to_file"))
        log_file_name = self.config.get_option("logging.log_file_name") + ".log"
        enable_logging = str2bool(self.config.get_option("logging.enable_logging"))

        logger = loguru_logger
        logger.remove()
        if enable_logging:
            logger.add(sys.stdout, level=logging_level)
            if log_to_file:
                logger.add(log_file_name, level=logging_level, rotation="100 MB")
        self._logger = logger



from relevanceai.base import _Base
from relevanceai.api.endpoints.centroids import CentroidsClient


class ClusterClient(_Base):
    def __init__(self, project, api_key):
        self.project = project
        self.api_key = api_key
        self.centroids = CentroidsClient(project=project, api_key=api_key)
        super().__init__(project, api_key)

    def aggregate(
        self,
        dataset_id: str,
        vector_field: str,
        metrics: list = [],
        groupby: list = [],
        filters: list = [],
        page_size: int = 20,
        page: int = 1,
        asc: bool = False,
        flatten: bool = True,
        alias: str = "default",
    ):
        """
        Takes an aggregation query and gets the aggregate of each cluster in a collection. This helps you interpret each cluster and what is in them.
        It can only can be used after a vector field has been clustered. \n

        For more information about aggregations check out services.aggregate.aggregate.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        vector_field : string
            The vector field that was clustered on
        metrics: list
            Fields and metrics you want to calculate
        groupby: list
            Fields you want to split the data into
        filters: list
            Query for filtering the search results
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order
        flatten: bool
            Whether to flatten
        alias: string
            Alias used to name a vector field. Belongs in field_{alias}vector
        """
        endpoint = "/services/cluster/aggregate"
        method = "POST"
        parameters = {
            "dataset_id": dataset_id,
            "aggregation_query": {"groupby": groupby, "metrics": metrics},
            "filters": filters,
            "page_size": page_size,
            "page": page,
            "asc": asc,
            "flatten": flatten,
            "vector_field": vector_field,
            "alias": alias,
        }
        self._log_to_dashboard(
            method=method,
            parameters=parameters,
            endpoint=endpoint,
            dashboard_type="cluster_aggregation",
        )
        return self.make_http_request(
            endpoint=endpoint, method=method, parameters=parameters
        )

    def facets(
        self,
        dataset_id: str,
        facets_fields: list = [],
        page_size: int = 20,
        page: int = 1,
        asc: bool = False,
        date_interval: str = "monthly",
    ):
        """
        Takes a high level aggregation of every field and every cluster in a collection. This helps you interpret each cluster and what is in them. \n
        Only can be used after a vector field has been clustered.

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        facets_fields : list
            Fields to include in the facets, if [] then all
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        asc: bool
            Whether to sort results by ascending or descending order
        date_interval: string
            Interval for date facets
        """
        return self.make_http_request(
            endpoint="/services/cluster/facets",
            method="GET",
            parameters={
                "dataset_id": dataset_id,
                "facets_fields": facets_fields,
                "page_size": page_size,
                "page": page,
                "asc": asc,
                "date_interval": date_interval,
            },
        )



"""Missing field error
"""


class RelevanceAIError(Exception):
    """Base class for all errors"""


class MissingFieldError(RelevanceAIError):
    """Error handling for missing fields"""


class APIError(RelevanceAIError):
    """Error related to API"""


class ClusteringResultsAlreadyExistsError(RelevanceAIError):
    """Exception raised for existing clustering results

    Attributes:
        message -- explanation of the error
    """

    def __init__(
        self, field_name, message="""Clustering results for %s already exist"""
    ):
        self.field_name = field_name
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return self.message % (self.field_name)



import dash
from dash import dcc
from dash import html
import plotly.graph_objs as go
from relevanceai.visualise.dash_components.components.sections import Card


def build_graph(app: dash.Dash, data, layout) -> html.Div:
    """
    Builds the graph component of the layout.
    """
    return html.Div(
        className="six columns",
        children=[
            dcc.Graph(
                figure=go.Figure(data=data, layout=layout),
                id="graph-plot-tsne",
                style={"height": "100vh"},
                config={"displayModeBar": False},
            ),
        ],
    )



from dash.dependencies import Input, Output
import plotly.graph_objs as go
import dash
from dash import dcc
import dash_bootstrap_components as dbc

import plotly.express as px
from skimage import io
from plotly.subplots import make_subplots

from relevanceai.vector_tools.nearest_neighbours import NearestNeighbours
from doc_utils.doc_utils import DocUtils

doc_utils = DocUtils()

MAX_SIZE = 200


def display_callbacks(app, show_image, docs, vector_label):

    if show_image:

        @app.callback(
            Output("div-plot-click-image", "children"),
            Input("graph-plot-tsne", "clickData"),
        )
        def display_image(clickData, height=100):
            try:
                click_id = clickData["points"][0]["customdata"][0]
            except TypeError:
                return None

            click_doc = [i for i in docs if i["_id"] == click_id][0]
            image_url = click_doc[vector_label]

            return [
                dash.html.P("currently selected item"),
                dash.html.Img(src=image_url, height=height),
            ]

    else:

        @app.callback(
            Output("div-plot-click-message", "children"),
            Input("graph-plot-tsne", "clickData"),
        )
        def display_text(clickData):
            try:
                return f"Current Selection: {clickData['points'][0]['customdata'][1]}"
            except TypeError:
                return None


def neighbour_callbacks(
    app,
    show_image,
    docs,
    vector_label,
    vector_field,
    distance_measure_mode="cosine",
    n=11,
):
    def _get_neighbours(clickData):
        try:
            click_id = clickData["points"][0]["customdata"][0]
        except TypeError:
            return None

        click_doc = [i for i in docs if i["_id"] == click_id][0]

        click_vec = click_doc[vector_field]
        nearest_neighbors = NearestNeighbours.get_nearest_neighbours(
            docs, click_vec, vector_field, distance_measure_mode
        )[:n]
        nearest_neighbor_values = doc_utils.get_field_across_documents(
            vector_label, nearest_neighbors
        )
        nearest_neighbor_index = doc_utils.get_field_across_documents(
            "nearest_neighbour_distance", nearest_neighbors
        )

        nearest_neighbor_index = [round(i, 2) for i in nearest_neighbor_index]

        nearest_neighbor_values, nearest_neighbor_index = remove_duplicates(
            nearest_neighbor_values, nearest_neighbor_index
        )

        if distance_measure_mode != "cosine":
            nearest_neighbor_index = nearest_neighbor_index[::-1]
            nearest_neighbor_values = nearest_neighbor_values[::-1]

        return {
            "nearest_neighbor_values": nearest_neighbor_values,
            "nearest_neighbor_index": nearest_neighbor_index,
        }

    if show_image:

        @app.callback(
            Output("div-plot-image-neighbours", "children"),
            Input("graph-plot-tsne", "clickData"),
        )
        def image_neighbours_html(clickData):
            neighbour_info = _get_neighbours(clickData)
            image_width = 50
            image_height = 50
            if neighbour_info:
                COLS = 5

                def generate_card(img=None, score: float = None, rank: int = None):
                    if rank == 0:
                        message = "Currently selected item"
                    else:
                        message = f"rank: {rank} | score: {score}"
                    card = dbc.Card(
                        dbc.CardBody(
                            [
                                dash.html.P(message),
                                dash.html.Img(
                                    src=img,
                                    id="card-value",
                                    width=image_width,
                                    height=image_height,
                                ),
                            ]
                        ),
                    )
                    return card

                # Attempt to create a grid layout
                layout = [dash.html.H5(f"Nearest Neighbours ({distance_measure_mode})")]
                layout_row = []
                for n, image in enumerate(neighbour_info["nearest_neighbor_values"]):
                    card = generate_card(
                        img=image,
                        rank=n,
                        score=neighbour_info["nearest_neighbor_index"][n],
                    )
                    col = dbc.Col([card])
                    layout_row.append(col)
                    if n % 2 == 1:
                        rows = dbc.Row(layout_row.copy(), className="mb-4")
                        layout.append(rows)
                        del layout_row
                        layout_row = []
                return layout

    else:

        @app.callback(
            Output("div-plot-text-neighbours", "children"),
            Input("graph-plot-tsne", "clickData"),
        )
        def text_neighbours(clickData):

            # Need to remove duplicates later
            neighbour_info = _get_neighbours(clickData)

            if neighbour_info:

                fig = px.bar(
                    x=neighbour_info["nearest_neighbor_index"][::-1],
                    y=neighbour_info["nearest_neighbor_values"][::-1],
                    color=neighbour_info["nearest_neighbor_index"][::-1],
                    text=neighbour_info["nearest_neighbor_index"][::-1],
                    color_continuous_scale="blues",
                    orientation="h",
                    title=f"Nearest Neighbours ({distance_measure_mode})",
                )
                fig.update_yaxes(title="")
                fig.update_xaxes(visible=False)
                fig.update_coloraxes(showscale=False)
                fig.update_traces(
                    hoverinfo="skip",
                    hovertemplate=None,
                    textposition="outside",
                    cliponaxis=False,
                    width=0.2,
                )
                fig.update_layout(
                    {"plot_bgcolor": "#ffffff", "paper_bgcolor": "#ffffff"}
                )

                return dcc.Graph(
                    figure=fig,
                    style={
                        "height": "125vh",
                        "width": "100vh",
                        "text-align": "center",
                    },
                    config={"displayModeBar": False},
                )

            return None


def update_graph():
    pass


def remove_duplicates(value, index):
    temp_dict = {i: j for i, j in zip(value, index)}
    return list(temp_dict.keys()), list(temp_dict.values())



"""The Transport Class defines a transport as used by the Channel class to communicate with the network.
"""
import time
import traceback
import json
from typing import Union
from json.decoder import JSONDecodeError

from urllib.parse import urlparse

import requests
from requests import Request

from relevanceai.config import Config
from relevanceai.logger import AbstractLogger
from relevanceai.dashboard_mappings import DASHBOARD_MAPPINGS
from relevanceai.errors import APIError


class Transport:
    """Base class for all relevanceai objects"""

    project: str
    api_key: str
    config: Config
    logger: AbstractLogger

    @property
    def _dashboard_request_url(self):
        return self.config.get_option("dashboard.dashboard_request_url")[1:-1]

    @property
    def auth_header(self):
        return {"Authorization": self.project + ":" + self.api_key}

    @property
    def _search_dashboard_url(self):
        return (
            self.config["dashboard.base_dashboard_url"][1:-1]
            + self.config["dashboard.search_dashboard_endpoint"][1:-1]
        )

    @staticmethod
    def _is_search_in_path(url: str):
        if url is None:
            return False
        result = urlparse(url)
        return "search" in result.path.split("/")

    @property
    def DASHBOARD_TYPES(self):
        return list(DASHBOARD_MAPPINGS.keys())

    def _log_to_dashboard(
        self,
        method: str,
        parameters: dict,
        endpoint: str,
        dashboard_type: str,
        verbose: bool = True,
    ):
        """Log search to dashboard"""
        # Needs to be a supported dashboard type
        if dashboard_type not in self.DASHBOARD_TYPES:
            return
        url = self.config.get_option("api.base_url")[:-2]
        version = self.config.get_option("api.base_url")[-2:]
        request_body = {
            dashboard_type: {
                "body": parameters,
                "url": url,
                "version": version,
                "endpoint": endpoint[1:],
                "metadata": parameters,
                "query": parameters.get("query"),
            },
        }
        req = Request(
            method=method.upper(),
            url=self._dashboard_request_url,
            headers=self.auth_header,
            json=request_body,
            # params=parameters if method.upper() == "GET" else {},
        ).prepare()
        with requests.Session() as s:
            response = s.send(req)

        if verbose:
            dashboard_url = (
                self.config["dashboard.base_dashboard_url"][1:-1]
                + DASHBOARD_MAPPINGS[dashboard_type]
            )
            self.print_dashboard_url(dashboard_url)
        return response

    def _log_search_to_dashboard(self, method: str, parameters: dict, endpoint: str):
        """Log search to dashboard"""
        return self._log_to_dashboard(
            method=method,
            parameters=parameters,
            endpoint=endpoint,
            dashboard_type="multivector_search",
        )

    def print_dashboard_url(self, dashboard_url):
        print(f"You can now visit the dashboard at {dashboard_url}")

    def make_http_request(
        self,
        endpoint: str,
        method: str = "GET",
        parameters: dict = {},
        base_url: str = None,
        output_format=None,
    ):
        """
        Make the HTTP request
        Parameters
        ----------
        endpoint: string
            The endpoint from the documentation to use
        method_type: string
            POST or GET request
        """
        self._last_used_endpoint = endpoint
        start_time = time.perf_counter()

        if base_url is None:
            # if Transport.is_search_in_path(base_url) and not hasattr(self, "output_format"):
            #     base_url = self.config.get_option("dashboard.base_dashboard_url")[1:-1]
            # else:
            base_url = self.config.get_option("api.base_url")

        if output_format is None:
            output_format = self.config.get_option("api.output_format")

        retries = int(self.config.get_option("retries.number_of_retries"))
        seconds_between_retries = int(
            self.config.get_option("retries.seconds_between_retries")
        )
        request_url = base_url + endpoint
        for _ in range(retries):

            self.logger.info("URL you are trying to access:" + request_url)
            try:
                if Transport._is_search_in_path(request_url):
                    self._log_search_to_dashboard(
                        method=method, parameters=parameters, endpoint=endpoint
                    )
                # TODO: Add other endpoints in here too

                req = Request(
                    method=method.upper(),
                    url=request_url,
                    headers=self.auth_header,
                    json=parameters if method.upper() == "POST" else {},
                    params=parameters if method.upper() == "GET" else {},
                ).prepare()

                with requests.Session() as s:
                    response = s.send(req)

                # Successful response
                if response.status_code == 200:
                    self._log_response_success(base_url, endpoint)
                    self._log_response_time(
                        base_url, endpoint, time.perf_counter() - start_time
                    )

                    if output_format == "json":
                        return response.json()
                    elif output_format == "content":
                        return response.content
                    elif output_format == "status_code":
                        return response.status_code
                    else:
                        return response

                # Cancel bad URLs
                elif response.status_code == 404:
                    self._log_response_fail(
                        base_url,
                        endpoint,
                        response.status_code,
                        response.content.decode(),
                    )
                    raise APIError(response.content.decode())

                # Retry other errors
                else:
                    self._log_response_fail(
                        base_url,
                        endpoint,
                        response.status_code,
                        response.content.decode(),
                    )
                    continue

            except (ConnectionError) as error:
                # Print the error
                traceback.print_exc()
                self._log_connection_error(base_url, endpoint)
                time.sleep(seconds_between_retries)
                continue

            except JSONDecodeError as error:
                self._log_no_json(base_url, endpoint, response.status_code, response)
                return response

        return response

    def _log_response_success(self, base_url, endpoint):
        self.logger.success(f"Response success! ({base_url + endpoint})")

    def _log_response_time(self, base_url, endpoint, time):
        self.logger.debug(f"Request ran in {time} seconds ({base_url + endpoint})")

    def _log_response_fail(self, base_url, endpoint, status_code, content):
        self.logger.error(
            f"Response failed ({base_url + endpoint}) (Status: {status_code} Response: {content})"
        )

    def _log_connection_error(self, base_url, endpoint):
        self.logger.error(f"Connection error but re-trying. ({base_url + endpoint})")

    def _log_no_json(self, base_url, endpoint, status_code, content):
        self.logger.error(
            f"No JSON Available ({base_url + endpoint}) (Status: {status_code} Response: {content})"
        )



import scipy.spatial.distance as spatial_distance
from relevanceai.base import _Base
from doc_utils.doc_utils import DocUtils
from relevanceai.vector_tools.constants import NEAREST_NEIGHBOURS

doc_utils = DocUtils()


class NearestNeighbours(_Base, DocUtils):
    def __init__(self, project: str, api_key: str):
        self.project = project
        self.api_key = api_key
        super().__init__(project, api_key)

    @staticmethod
    def get_nearest_neighbours(
        docs: list,
        vector: list,
        vector_field: str,
        distance_measure_mode: NEAREST_NEIGHBOURS = "cosine",
        callable_distance=None,
    ):

        if callable_distance:
            sort_key = [
                callable_distance(i, vector)
                for i in doc_utils.get_field_across_documents(vector_field, docs)
            ]
            reverse = False

        elif distance_measure_mode == "cosine":
            sort_key = [
                1 - spatial_distance.cosine(i, vector)
                for i in doc_utils.get_field_across_documents(vector_field, docs)
            ]
            reverse = True

        elif distance_measure_mode == "l2":
            sort_key = [
                spatial_distance.euclidean(i, vector)
                for i in doc_utils.get_field_across_documents(vector_field, docs)
            ]
            reverse = False

        else:
            raise ValueError("Need valid distance measure mode or callable distance")

        doc_utils.set_field_across_documents(
            "nearest_neighbour_distance", sort_key, docs
        )

        return sorted(
            docs, reverse=reverse, key=lambda x: x["nearest_neighbour_distance"]
        )



"""access the client via this class
"""
import getpass
import json
import os
import sys
import warnings
from typing import Optional, List, Union

from doc_utils.doc_utils import DocUtils

from relevanceai.errors import APIError
from relevanceai.api.client import BatchAPIClient
from relevanceai.api.endpoints.cluster import ClusterClient
from relevanceai.config import CONFIG
from relevanceai.vector_tools.cluster import KMeans

vis_requirements = False
try:
    from relevanceai.visualise.projector import Projector

    vis_requirements = True
except ModuleNotFoundError as e:
    warnings.warn(f"{e} You can fix this by installing RelevanceAI[vis]")
    pass

from relevanceai.vector_tools.client import VectorTools


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


class Client(BatchAPIClient, DocUtils):
    """Python Client for Relevance AI's relevanceai"""

    FAIL_MESSAGE = """Your API key is invalid. Please login again"""
    _cred_fn = ".creds.json"

    def __init__(
        self,
        project=os.getenv("RELEVANCE_PROJECT"),
        api_key=os.getenv("RELEVANCE_API_KEY"),
        authenticate: bool = False,
    ):

        if project is None or api_key is None:
            project, api_key = self._token_to_auth()

        super().__init__(project, api_key)

        if authenticate:
            if self.check_auth():

                WELCOME_MESSAGE = f"""Welcome to the RelevanceAI Python SDK. Logged in as {project}."""
                print(self.WELCOME_MESSAGE)
            else:
                raise APIError(self.FAIL_MESSAGE)

        if vis_requirements:
            self.projector = Projector(project, api_key)
        else:
            self.logger.warning(
                "Projector not loaded. You do not have visualisation requirements installed."
            )
        self.vector_tools = VectorTools(project, api_key)

    # @property
    # def output_format(self):
    #     return CONFIG.get_field("api.output_format", CONFIG.config)

    # @output_format.setter
    # def output_format(self, value):
    #     CONFIG.set_option("api.output_format", value)

    @property
    def base_url(self):
        return CONFIG.get_field("api.base_url", CONFIG.config)

    @base_url.setter
    def base_url(self, value):
        CONFIG.set_option("api.base_url", value)

    def _token_to_auth(self):
        # if verbose:
        #     print("You can sign up/login and find your credentials here: https://cloud.relevance.ai/sdk/api")
        #     print("Once you have signed up, click on the value under `Authorization token` and paste it here:")
        # SIGNUP_URL = "https://auth.relevance.ai/signup/?callback=https%3A%2F%2Fcloud.relevance.ai%2Flogin%3Fredirect%3Dcli-api"
        SIGNUP_URL = "https://cloud.relevance.ai/sdk/api"
        if not os.path.exists(self._cred_fn):
            # We repeat it twice because of different behaviours
            print(f"Authorization token (you can find it here: {SIGNUP_URL} )")
            token = getpass.getpass(
                f"Authorization token (you can find it here: {SIGNUP_URL} )"
            )
            project = token.split(":")[0]
            api_key = token.split(":")[1]
            self._write_credentials(project, api_key)
        else:
            data = self._read_credentials()
            project = data["project"]
            api_key = data["api_key"]
        return project, api_key

    def _write_credentials(self, project, api_key):
        json.dump({"project": project, "api_key": api_key}, open(self._cred_fn, "w"))

    def _read_credentials(self):
        return json.load(open(self._cred_fn))

    def login(
        self,
        authenticate: bool = True,
    ):
        """Preferred login method for demos and interactive usage."""
        project, api_key = self._token_to_auth()
        return Client(project=project, api_key=api_key, authenticate=authenticate)

    @property
    def auth_header(self):
        return {"Authorization": self.project + ":" + self.api_key}

    def make_search_suggestion(self):
        return self.services.search.make_suggestion()

    def check_auth(self):
        return self.admin._ping()



from relevanceai.base import _Base


class EncodersClient(_Base):
    def __init__(self, project: str, api_key: str):
        self.project = project
        self.api_key = api_key
        super().__init__(project, api_key)

    def textimage(self, text: str):
        """
        Encode text to make searchable with images

        Parameters
        ----------
        text: string
            Text to encode
        """
        return self.make_http_request(
            "/services/encoders/textimage", method="GET", parameters={"text": text}
        )

    def text(self, text: str):
        """
        Encode text

        Parameters
        ----------
        text: string
            Text to encode
        """
        return self.make_http_request(
            "/services/encoders/text", method="GET", parameters={"text": text}
        )

    def multi_text(self, text):
        """
        Encode multilingual text

        Parameters
        ----------
        text: string
            Text to encode
        """
        return self.make_http_request(
            "/services/encoders/multi_text", method="GET", parameters={"text": text}
        )

    def image(self, image):
        """
        Encode an image

        Parameters
        ----------
        image: string
            URL of image to encode
        """
        return self.make_http_request(
            "/services/encoders/image", method="POST", parameters={"image": image}
        )

    def imagetext(self, image):
        """
        Encode an image to make searchable with text

        Parameters
        ----------
        image: string
            URL of image to encode
        """
        return self.make_http_request(
            "/services/encoders/imagetext", method="GET", parameters={"image": image}
        )



"""Dashboard mappings
"""

DASHBOARD_MAPPINGS = {
    "multivector_search": "/sdk/search",
    "cluster_centroids_closest": "/sdk/cluster/centroids/closest",
    "cluster_centroids_furthest": "/sdk/cluster/centroids/furthest",
    "cluster_aggregation": "/sdk/cluster/aggregation",
}



"""Batch Insert"""
import json
import math
import sys
import time
import traceback
from datetime import datetime
from typing import Callable, List, Dict, Union, Any

from relevanceai.api.endpoints.client import APIClient
from relevanceai.api.batch.batch_retrieve import BatchRetrieveClient
from relevanceai.api.batch.local_logger import PullUpdatePushLocalLogger
from relevanceai.concurrency import multiprocess, multithread
from relevanceai.progress_bar import progress_bar
from relevanceai.api.batch.chunk import Chunker

BYTE_TO_MB = 1024 * 1024
LIST_SIZE_MULTIPLIER = 3


class BatchInsertClient(BatchRetrieveClient, APIClient, Chunker):
    def insert_documents(
        self,
        dataset_id: str,
        docs: list,
        bulk_fn: Callable = None,
        max_workers: int = 8,
        retry_chunk_mult: float = 0.5,
        show_progress_bar: bool = False,
        chunksize=0,
        *args,
        **kwargs,
    ):

        """
        Insert a list of documents with multi-threading automatically enabled.

        - When inserting the document you can optionally specify your own id for a document by using the field name "_id", if not specified a random id is assigned.
        - When inserting or specifying vectors in a document use the suffix (ends with) "_vector_" for the field name. e.g. "product_description_vector_".
        - When inserting or specifying chunks in a document the suffix (ends with) "_chunk_" for the field name. e.g. "products_chunk_".
        - When inserting or specifying chunk vectors in a document's chunks use the suffix (ends with) "_chunkvector_" for the field name. e.g. "products_chunk_.product_description_chunkvector_".

        Documentation can be found here: https://ingest-api-dev-aueast.relevance.ai/latest/documentation#operation/InsertEncode

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        docs : list
            A list of documents. Document is a JSON-like data that we store our metadata and vectors with. For specifying id of the document use the field '_id', for specifying vector field use the suffix of '_vector_'
        bulk_fn : callable
            Function to apply to documents before uploading
        max_workers : int
            Number of workers active for multi-threading
        retry_chunk_mult: int
            Multiplier to apply to chunksize if upload fails
        chunksize : int
            Number of documents to upload per worker. If None, it will default to the size specified in config.upload.target_chunk_mb
        """

        self.logger.info(f"You are currently inserting into {dataset_id}")

        self.logger.info(
            f"You can track your stats and progress via our dashboard at https://cloud.relevance.ai/collections/dashboard/stats/?collection={dataset_id}"
        )
        # Check if the collection exists
        self.datasets.create(dataset_id)

        def bulk_insert_func(docs):
            return self.datasets.bulk_insert(
                dataset_id,
                docs,
                return_documents=True,
                *args,
                **kwargs,
            )

        return self._write_documents(
            bulk_insert_func,
            docs,
            bulk_fn,
            max_workers,
            retry_chunk_mult,
            show_progress_bar=show_progress_bar,
            chunksize=chunksize,
        )

    def update_documents(
        self,
        dataset_id: str,
        docs: list,
        bulk_fn: Callable = None,
        max_workers: int = 8,
        retry_chunk_mult: float = 0.5,
        chunksize: int = 0,
        show_progress_bar=False,
        *args,
        **kwargs,
    ):
        """
        Update a list of documents with multi-threading automatically enabled.
        Edits documents by providing a key value pair of fields you are adding or changing, make sure to include the "_id" in the documents.

        >>> from relevanceai import Client
        >>> url = "https://api-aueast.relevance.ai/v1/"
        >>> collection = ""
        >>> project = ""
        >>> api_key = ""
        >>> client = Client(project, api_key)
        >>> docs = client.datasets.documents.get_where(collection, select_fields=['title'])
        >>> while len(docs['documents']) > 0:
        >>>     docs['documents'] = model.encode_documents_in_bulk(['product_name'], docs['documents'])
        >>>     client.update_documents(collection, docs['documents'])
        >>>     docs = client.datasets.documents.get_where(collection, select_fields=['product_name'], cursor=docs['cursor'])

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        docs : list
            A list of documents. Document is a JSON-like data that we store our metadata and vectors with. For specifying id of the document use the field '_id', for specifying vector field use the suffix of '_vector_'
        bulk_fn : callable
            Function to apply to documents before uploading
        max_workers : int
            Number of workers active for multi-threading
        retry_chunk_mult: int
            Multiplier to apply to chunksize if upload fails
        chunksize : int
            Number of documents to upload per worker. If None, it will default to the size specified in config.upload.target_chunk_mb
        """

        self.logger.info(f"You are currently updating {dataset_id}")

        self.logger.info(
            f"You can track your stats and progress via our dashboard at https://cloud.relevance.ai/collections/dashboard/stats/?collection={dataset_id}"
        )

        def bulk_update_func(docs):
            return self.datasets.documents.bulk_update(
                dataset_id,
                docs,
                return_documents=True,
                *args,
                **kwargs,
            )

        return self._write_documents(
            bulk_update_func,
            docs,
            bulk_fn,
            max_workers,
            retry_chunk_mult,
            chunksize=chunksize,
        )

    def pull_update_push(
        self,
        original_collection: str,
        update_function,
        updated_collection: str = None,
        log_file: str = None,
        updating_args: dict = {},
        retrieve_chunk_size: int = 100,
        max_workers: int = 8,
        filters: list = [],
        select_fields: list = [],
        show_progress_bar: bool = True,
    ):
        """
        Loops through every document in your collection and applies a function (that is specified by you) to the documents.
        These documents are then uploaded into either an updated collection, or back into the original collection.

        Parameters
        ----------
        original_collection : string
            The dataset_id of the collection where your original documents are
        logging_collection: string
            The dataset_id of the collection which logs which documents have been updated. If 'None', then one will be created for you.
        updated_collection: string
            The dataset_id of the collection where your updated documents are uploaded into. If 'None', then your original collection will be updated.
        update_function: function
            A function created by you that converts documents in your original collection into the updated documents. The function must contain a field which takes in a list of documents from the original collection. The output of the function must be a list of updated documents.
        updating_args: dict
            Additional arguments to your update_function, if they exist. They must be in the format of {'Argument': Value}
        retrieve_chunk_size: int
            The number of documents that are received from the original collection with each loop iteration.
        max_workers: int
            The number of processors you want to parallelize with
        max_error:
            How many failed uploads before the function breaks
        """
        if not callable(update_function):
            raise TypeError(
                "Your update function needs to be a function! Please read the documentation if it is not."
            )

        # Check if a logging_collection has been supplied
        if log_file is None:
            log_file = (
                original_collection
                + "_"
                + str(datetime.now().strftime("%d-%m-%Y-%H-%M-%S"))
                + "_pull_update_push"
                + ".log"
            )

        # Instantiate the logger to document the successful IDs
        PULL_UPDATE_PUSH_LOGGER = PullUpdatePushLocalLogger(log_file)

        # Track failed documents
        failed_documents: List[Dict] = []

        # Trust the process
        # Get document lengths to calculate iterations
        original_length = self.get_number_of_documents(original_collection, filters)

        # get the remaining number in case things break
        remaining_length = original_length - PULL_UPDATE_PUSH_LOGGER.count_ids_in_fn()
        iterations_required = math.ceil(remaining_length / retrieve_chunk_size)

        completed_documents_list: list = []

        # Get incomplete documents from raw collection
        retrieve_filters = filters + [
            {
                "field": "ids",
                "filter_type": "ids",
                "condition": "!=",
                "condition_value": completed_documents_list,
            }
        ]

        for _ in progress_bar(
            range(iterations_required), show_progress_bar=show_progress_bar
        ):

            orig_json = self.datasets.documents.get_where(
                original_collection,
                filters=retrieve_filters,
                page_size=retrieve_chunk_size,
                select_fields=select_fields,
            )

            documents = orig_json["documents"]

            try:
                updated_data = update_function(documents, **updating_args)
            except Exception as e:
                self.logger.error("Your updating function does not work: " + str(e))
                traceback.print_exc()
                return

            updated_documents = [i["_id"] for i in documents]

            # Upload documents
            if updated_collection is None:
                insert_json = self.update_documents(
                    dataset_id=original_collection,
                    docs=updated_data,
                    max_workers=max_workers,
                    show_progress_bar=False,
                )
            else:
                insert_json = self.insert_documents(
                    dataset_id=updated_collection,
                    docs=updated_data,
                    max_workers=max_workers,
                    show_progress_bar=False,
                )

            # Check success
            chunk_failed = insert_json["failed_documents"]
            failed_documents.extend(chunk_failed)
            success_documents = list(set(updated_documents) - set(failed_documents))
            PULL_UPDATE_PUSH_LOGGER.log_ids(success_documents)
            self.logger.success(
                f"Chunk of {retrieve_chunk_size} original documents updated and uploaded with {len(chunk_failed)} failed documents!"
            )

        self.logger.success(f"Pull, Update, Push is complete!")

        # if PULL_UPDATE_PUSH_LOGGER.count_ids_in_fn() == original_length:
        #     os.remove(log_file)
        return {
            "failed_documents": failed_documents,
        }

    def pull_update_push_to_cloud(
        self,
        original_collection: str,
        update_function,
        updated_collection: str = None,
        logging_collection: str = None,
        updating_args: dict = {},
        retrieve_chunk_size: int = 100,
        retrieve_chunk_size_failure_retry_multiplier: float = 0.5,
        number_of_retrieve_retries: int = 3,
        max_workers: int = 8,
        max_error: int = 1000,
        filters: list = [],
        select_fields: list = [],
        show_progress_bar: bool = True,
    ):
        """
        Loops through every document in your collection and applies a function (that is specified by you) to the documents.
        These documents are then uploaded into either an updated collection, or back into the original collection.

        Parameters
        ----------
        original_collection : string
            The dataset_id of the collection where your original documents are
        logging_collection: string
            The dataset_id of the collection which logs which documents have been updated. If 'None', then one will be created for you.
        updated_collection: string
            The dataset_id of the collection where your updated documents are uploaded into. If 'None', then your original collection will be updated.
        update_function: function
            A function created by you that converts documents in your original collection into the updated documents. The function must contain a field which takes in a list of documents from the original collection. The output of the function must be a list of updated documents.
        updating_args: dict
            Additional arguments to your update_function, if they exist. They must be in the format of {'Argument': Value}
        retrieve_chunk_size: int
            The number of documents that are received from the original collection with each loop iteration.
        retrieve_chunk_size_failure_retry_multiplier: int
            If fails, retry on each chunk
        max_workers: int
            The number of processors you want to parallelize with
        max_error:
            How many failed uploads before the function breaks
        """
        # Check if a logging_collection has been supplied
        if logging_collection is None:
            logging_collection = (
                original_collection
                + "_"
                + str(datetime.now().strftime("%d-%m-%Y-%H-%M-%S"))
                + "_pull_update_push"
            )

        # Check collections and create completed list if needed
        collection_list = self.datasets.list()
        if logging_collection not in collection_list["datasets"]:
            self.logger.info("Creating a logging collection for you.")
            self.logger.info(self.datasets.create(logging_collection))

        # Track failed documents
        failed_documents: List[Dict] = []

        # Trust the process
        for _ in range(number_of_retrieve_retries):

            # Get document lengths to calculate iterations
            original_length = self.get_number_of_documents(original_collection, filters)
            completed_length = self.get_number_of_documents(logging_collection)
            remaining_length = original_length - completed_length
            iterations_required = math.ceil(remaining_length / retrieve_chunk_size)

            self.logger.debug(f"{original_length}")
            self.logger.debug(f"{completed_length}")
            self.logger.debug(f"{iterations_required}")

            # Return if no documents to update
            if remaining_length == 0:
                self.logger.success(f"Pull, Update, Push is complete!")
                return {
                    "failed_documents": failed_documents,
                    "logging_collection": logging_collection,
                }

            for _ in progress_bar(
                range(iterations_required), show_progress_bar=show_progress_bar
            ):

                # Get completed documents
                log_json = self.get_all_documents(logging_collection)
                completed_documents_list = [i["_id"] for i in log_json]

                # Get incomplete documents from raw collection
                retrieve_filters = filters + [
                    {
                        "field": "ids",
                        "filter_type": "ids",
                        "condition": "!=",
                        "condition_value": completed_documents_list,
                    }
                ]

                orig_json = self.datasets.documents.get_where(
                    original_collection,
                    filters=retrieve_filters,
                    page_size=retrieve_chunk_size,
                    select_fields=select_fields,
                )

                documents = orig_json["documents"]
                self.logger.debug(f"{len(documents)}")

                # Update documents
                try:
                    updated_data = update_function(documents, **updating_args)
                except Exception as e:
                    self.logger.error("Your updating function does not work: " + str(e))
                    traceback.print_exc()
                    return
                updated_documents = [i["_id"] for i in documents]
                self.logger.debug(f"{len(updated_data)}")

                # Upload documents
                if updated_collection is None:
                    insert_json = self.update_documents(
                        dataset_id=original_collection,
                        docs=updated_data,
                        max_workers=max_workers,
                        show_progress_bar=False,
                    )
                else:
                    insert_json = self.insert_documents(
                        dataset_id=updated_collection,
                        docs=updated_data,
                        max_workers=max_workers,
                        show_progress_bar=False,
                    )

                # Check success
                chunk_failed = insert_json["failed_documents"]
                self.logger.success(
                    f"Chunk of {retrieve_chunk_size} original documents updated and uploaded with {len(chunk_failed)} failed documents!"
                )
                failed_documents.extend(chunk_failed)
                success_documents = list(set(updated_documents) - set(failed_documents))
                upload_documents = [{"_id": i} for i in success_documents]

                self.insert_documents(
                    logging_collection,
                    upload_documents,
                    max_workers=max_workers,
                )

                # If fail, try to reduce retrieve chunk
                if len(chunk_failed) > 0:
                    self.logger.warning(
                        "Failed to upload. Retrieving half of previous number."
                    )
                    retrieve_chunk_size = int(
                        retrieve_chunk_size
                        * retrieve_chunk_size_failure_retry_multiplier
                    )
                    time.sleep(self.config.seconds_between_retries)
                    break

                if len(failed_documents) > max_error:
                    self.logger.error(
                        f"You have over {max_error} failed documents which failed to upload!"
                    )
                    return {
                        "failed_documents": failed_documents,
                        "logging_collection": logging_collection,
                    }

        self.logger.success(f"Pull, Update, Push is complete!")
        return {
            "failed_documents": failed_documents,
            "logging_collection": logging_collection,
        }

    def insert_df(self, dataset_id, dataframe, *args, **kwargs):
        """Insert a dataframe for eachd doc"""
        import pandas as pd

        docs = [
            {k: v for k, v in doc.items() if not pd.isna(v)}
            for doc in dataframe.to_dict(orient="records")
        ]
        return self.insert_documents(dataset_id, docs, *args, **kwargs)

    def delete_pull_update_push_logs(self, dataset_id=False):

        collection_list = self.datasets.list()["datasets"]

        if dataset_id:
            log_collections = [
                i
                for i in collection_list
                if ("pull_update_push" in i) and (dataset_id in i)
            ]

        else:
            log_collections = [i for i in collection_list if ("pull_update_push" in i)]

        [self.datasets.delete(i, confirm=False) for i in log_collections]
        return

    def _write_documents(
        self,
        insert_function,
        docs: list,
        bulk_fn: Callable = None,
        max_workers: int = 8,
        retry_chunk_mult: float = 0.5,
        show_progress_bar: bool = False,
        chunksize: int = 0,
    ):

        # Get one document to test the size
        if len(docs) == 0:
            self.logger.warning("No document is detected")
            return {
                "inserted": 0,
                "failed_documents": [],
                "failed_documents_detailed": [],
            }

        # Insert documents
        test_doc = json.dumps(docs[0], indent=4)
        doc_mb = sys.getsizeof(test_doc) * LIST_SIZE_MULTIPLIER / BYTE_TO_MB
        if chunksize == 0:
            target_chunk_mb = int(self.config.get_option("upload.target_chunk_mb"))
            chunksize = (
                int(target_chunk_mb / doc_mb) + 1
                if int(target_chunk_mb / doc_mb) + 1 < len(docs)
                else len(docs)
            )

        # Initialise number of inserted documents
        inserted: List[str] = []

        # Initialise failed documents
        failed_ids = [i["_id"] for i in docs]

        # Initialise failed documents detailed
        failed_ids_detailed: List[str] = []

        # Initialise cancelled documents
        cancelled_ids = []

        for i in range(int(self.config.get_option("retries.number_of_retries"))):
            if len(failed_ids) > 0:
                if bulk_fn is not None:
                    insert_json = multiprocess(
                        func=bulk_fn,
                        iterables=docs,
                        post_func_hook=insert_function,
                        max_workers=max_workers,
                        chunksize=chunksize,
                        show_progress_bar=show_progress_bar,
                    )
                else:
                    insert_json = multithread(
                        insert_function,
                        docs,
                        max_workers=max_workers,
                        chunksize=chunksize,
                        show_progress_bar=show_progress_bar,
                    )

                failed_ids = []
                failed_ids_detailed = []

                # Update inserted amount
                def is_successfully_inserted(chunk: Union[Dict, Any]) -> bool:
                    return chunk["status_code"] == 200

                inserted += list(
                    map(
                        lambda x: x["response_json"]["inserted"],
                        filter(is_successfully_inserted, insert_json),
                    )
                )

                for chunk in insert_json:

                    # Track failed in 200
                    if chunk["status_code"] == 200:
                        failed_ids += [
                            i["_id"] for i in chunk["response_json"]["failed_documents"]
                        ]

                        failed_ids_detailed += [
                            i for i in chunk["response_json"]["failed_documents"]
                        ]

                    # Cancel documents with 400 or 404
                    elif chunk["status_code"] in [400, 404]:
                        cancelled_ids += [i["_id"] for i in chunk["documents"]]

                    # Half chunksize with 413 or 524
                    elif chunk["status_code"] in [413, 524]:
                        failed_ids += [i["_id"] for i in chunk["documents"]]
                        chunksize = int(chunksize * retry_chunk_mult)

                    # Retry all other errors
                    else:
                        failed_ids += [i["_id"] for i in chunk["documents"]]

                # Update docs to retry which have failed
                docs = [i for i in docs if i["_id"] in failed_ids]

            else:
                break

        # When returning, add in the cancelled ids
        failed_ids.extend(cancelled_ids)

        output = {
            "inserted": sum(inserted),
            "failed_documents": failed_ids,
            "failed_documents_detailed": failed_ids_detailed,
        }
        return output



from typing import List, Sequence, Union, Dict, Any, Tuple, Optional
from typing_extensions import Literal

from joblib.memory import Memory

dict = Dict[str, Any]

DIM_REDUCTION = Literal["pca", "tsne", "umap", "ivis"]
DIM_REDUCTION_DEFAULT_ARGS: Dict[Any, Any] = {
    "pca": {"svd_solver": "auto", "random_state": 42},
    "tsne": {
        "init": "pca",
        "n_iter": 500,
        "learning_rate": 100,
        "perplexity": 30,
        "random_state": 42,
    },
    "umap": {
        "n_neighbors": 10,
        "min_dist": 0.1,
        "random_state": 42,
        "transform_seed": 42,
    },
    "ivis": {"k": 15, "model": "maaten", "n_epochs_without_progress": 2},
}


# CLUSTER = Literal["kmeans", "kmedoids", "kmodes", "kprototypes", None]
CLUSTER = Literal["kmeans", "kmedoids", "hdbscan"]
CLUSTER_DEFAULT_ARGS: Dict[Any, Any] = {
    "kmeans": {
        "init": "k-means++",
        "verbose": 0,
        # "compute_labels": True,
        # "max_no_improvement": 2,
    },
    "kmedoids": {
        "metric": "euclidean",
        "init": "k-medoids++",
        "random_state": 42,
        "method": "pam",
    },
    # 'kmodes': {
    #     "init": "Huang",
    #     "verbose": 0,
    #     "random_state": 42,
    #     "n_jobs": -1
    # },
    # 'kprototypes': {
    #     "init": "Huang",
    #     "verbose": 0,
    #     "random_state": 42,
    #     "n_jobs": -1
    # },
    "hdbscan": {
        "algorithm": "best",
        "alpha": 1.0,
        "approx_min_span_tree": True,
        "gen_min_span_tree": False,
        "leaf_size": 40,
        "memory": Memory(cachedir=None),
        "metric": "euclidean",
        "min_samples": None,
        "p": None,
    },
}

NEAREST_NEIGHBOURS = Literal["cosine", "l2"]



import dash
from dash import html
from relevanceai.visualise.dash_components.sections.header import build_header
from relevanceai.visualise.dash_components.sections.display_panel import (
    build_display_panel,
)
from relevanceai.visualise.dash_components.sections.graph import build_graph
from relevanceai.visualise.dash_components.callbacks import (
    display_callbacks,
    neighbour_callbacks,
)
from jupyter_dash import JupyterDash


def create_dash_graph(
    plot_data,
    layout,
    show_image,
    docs,
    vector_label,
    vector_field,
    interactive: bool = True,
):

    app = JupyterDash(__name__)

    def create_layout(app: dash.Dash) -> html.Div:
        """
        Create the layout of the Dash app.
        """
        return html.Div(
            className="row",
            style={"max-width": "100%", "font-size": "1.5rem", "padding": "0px 0px"},
            children=[
                ## --- Header --- ##
                build_header(app),
                html.Div(id="my-output"),
                ## --- Body --- ##
                html.Div(
                    className="row background",
                    style={"padding": "0px"},
                    children=[
                        build_graph(app, plot_data, layout),
                        build_display_panel(app, show_image=show_image),
                    ],
                ),
            ],
        )

    app.layout = create_layout(app)
    # display_callbacks(app, show_image, docs, vector_label)
    if interactive:
        neighbour_callbacks(app, show_image, docs, vector_label, vector_field)
    app.run_server(mode="inline")



"""Services class
"""
from relevanceai.base import _Base

from relevanceai.api.endpoints.encoders import EncodersClient
from relevanceai.api.endpoints.cluster import ClusterClient
from relevanceai.api.endpoints.search import SearchClient
from relevanceai.api.endpoints.aggregate import AggregateClient
from relevanceai.api.endpoints.recommend import RecommendClient
from relevanceai.api.endpoints.tagger import TaggerClient
from relevanceai.api.endpoints.prediction import PredictionClient
from relevanceai.api.endpoints.wordclouds import WordcloudsClient


class ServicesClient(_Base):
    def __init__(self, project: str, api_key: str):
        self.project = project
        self.api_key = api_key
        self.encoders = EncodersClient(project=project, api_key=api_key)
        self.cluster = ClusterClient(project=project, api_key=api_key)
        self.search = SearchClient(project=project, api_key=api_key)
        self.aggregate = AggregateClient(project=project, api_key=api_key)
        self.recommend = RecommendClient(project=project, api_key=api_key)
        self.tagger = TaggerClient(project=project, api_key=api_key)
        self.prediction = PredictionClient(project=project, api_key=api_key)
        self.wordclouds = WordcloudsClient(project=project, api_key=api_key)
        super().__init__(project, api_key)

    def document_diff(
        self, doc: dict, docs_to_compare: list, difference_fields: list = []
    ):
        """
        Find differences between documents

        Parameters
        ----------
        doc: dict
            Main document to compare other documents against.
        docs_to_compare: list
            Other documents to compare against the main document.
        difference_fields: list
            Fields to compare. Defaults to [], which compares all fields.

        """
        return self.make_http_request(
            endpoint=f"/services/document_diff",
            method="POST",
            parameters={
                "doc": doc,
                "docs_to_compare": docs_to_compare,
                "difference_fields": difference_fields,
            },
        )



"""Configuration Settings"""
import configparser
import os

from doc_utils.doc_utils import DocUtils

PATH = os.path.dirname(os.path.abspath(__file__))
CONFIG_PATH = os.path.join(PATH, "config.ini")


class Config(DocUtils):
    """
    Set and change configuration settings

    - Retries - Set the behaviour of retries for failed responses from the API
        - number_of_retries - Number of retries to attempt
        - seconds_between_retries - Seconds to wait between retries

    - Logging - Set the behaviour of logging
         - enable_logging - Whether to log
         - log_to_file - Whether to log to file
         - log_file_name - The name of the file to log to, if logging to file
         - logging_level - Minimum level to log

    - Upload - Set the behaviour of uploads to RelevanceAI
        - target_chunk_mb - Maximum upload size per request

    - API - Set the behaviour of API requests
        - base_url - The base url to access
        - output_format - The format of API responses

    - Dashboard - URLS to various things

    """

    def __init__(self):
        self.config = configparser.ConfigParser()
        self._read_config(CONFIG_PATH)
        super().__init__()

    def _read_config(self, config_path):
        """
        Set up custom config by reading in a file

        Parameters
        ----------
        config_path : string
            Path to config
        """
        self.config.read(config_path)

    @property
    def options(self):
        """View all current config settings"""
        return self.config._sections

    def get_option(self, option):
        """
        View current config settings

        Parameters
        ----------
        option : string
            Setting key
        """
        return self.get_field(option, self.config)

    def set_option(self, option, value):
        """
        Change a config settings

        Parameters
        ----------
        option : string
            Setting key
        value : string
            New setting
        """
        self.set_field(option, self.config, str(value))

    def reset_to_default(self):
        """Reset config to default"""
        self._read_config(CONFIG_PATH)

    @staticmethod
    def _create_default():
        config = configparser.ConfigParser()
        config["retries"] = {"number_of_retries": 3, "seconds_between_retries": 2}
        config["logging"] = {
            "log_to_file": False,
            "logging_level": "ERROR",
            "enable_logging": True,
            "log_file_name": "relevanceai",
        }
        config["upload"] = {"target_chunk_mb": 100}
        config["api"] = {
            "base_url": "https://gateway-api-aueast.relevance.ai/v1",
            "output_format": "json",
        }
        with open(CONFIG_PATH, "w") as configfile:
            config.write(configfile)

    def __getitem__(self, key):
        """
        Get teh config using client.config["api.base_url"]
        """
        return self.get_option(key)

    def __setitem__(self, key: str, value: str):
        """
        Set the config using client.config["api.base_url"] = "https://..."
        """
        return self.set_option(key, value)


CONFIG = Config()

# To create the initial config
if __name__ == "__main__":
    Config._create_default()
